﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;

namespace DemoLINQCS
{
    public partial class Form1 : Form
    {

        private List<Customer> _customers = new List<Customer>();
        private List<Product> _products = new List<Product>();
        private List<Product> _products2 = new List<Product>(); 

        public Form1()
        {
            InitializeComponent();
            LoadData();
        }

        void LoadData()
        {
            _customers.Add(new Customer {
                    Name = "Dalton, Joe", 
                    City = "Toronto", 
                    Country = Countries.Canada, 
                    Orders = new Order[] {
                        new Order {IdOrder = 1, Quantity = 3, IdProduct = 1, Shipped = false, Month = "January"},
                        new Order {IdOrder = 2, Quantity = 5, IdProduct = 2, Shipped = true, Month = "May"}}});
            _customers.Add(new Customer {
                    Name = "Dalton, Jack", 
                    City = "Montréal", 
                    Country = Countries.Canada, 
                    Orders =  new Order[] {
                        new Order {IdOrder = 3, Quantity = 10, IdProduct = 1, Shipped = false, Month = "July"},
                        new Order {IdOrder = 4, Quantity = 20, IdProduct = 3, Shipped = true, Month = "December"}}});
            _customers.Add(new Customer {
                    Name = "Dalton, William", 
                    City = "Winnipeg", 
                    Country = Countries.Canada, 
                    Orders =  new Order[] {
                        new Order {IdOrder = 5, Quantity = 20, IdProduct = 3, Shipped = true, Month = "December"}}});
            _customers.Add(new Customer {
                    Name = "Dalton, Averell", 
                    City = "Vancouver", 
                    Country = Countries.Canada,
                    Orders = new Order[] {
                        new Order {IdOrder = 6, Quantity = 20, IdProduct = 5, Shipped = false, Month = "July"}}});
            _customers.Add(new Customer{
                Name = "Moreau, Éric",
                City = "Montréal",
                Country = Countries.Canada,
                Orders = new Order[] {
                        new Order {IdOrder = 7, Quantity = 33, IdProduct = 1, Shipped = true, Month = "Juyly"},
                        new Order {IdOrder = 8, Quantity = 55, IdProduct = 2, Shipped = false, Month = "August"}}});
            _customers.Add(new Customer
            {
                Name = "US Customer 1",
                City = "New York",
                Country = Countries.USA,
                Orders = new Order[] {
                        new Order {IdOrder = 9, Quantity = 7, IdProduct = 5, Shipped = true, Month = "September"},
                        new Order {IdOrder = 10, Quantity = 4, IdProduct = 44, Shipped = false, Month = "October"}}
            });

            _products.Add(new Product { IdProduct = 1, Description = "Product 1", Price = 10 });
            _products.Add(new Product { IdProduct = 2, Description = "Product 2", Price = 20 });
            _products.Add(new Product { IdProduct = 3, Description = "Product 3", Price = 30 });
            _products.Add(new Product { IdProduct = 4, Description = "Product 4", Price = 40 });
            _products.Add(new Product { IdProduct = 5, Description = "Product 5", Price = 50 });
            _products.Add(new Product { IdProduct = 6, Description = "Product 6", Price = 60 });

            _products2.Add(new Product { IdProduct = 1, Description = "Product 1", Price = 10 });
            _products2.Add(new Product { IdProduct = 3, Description = "Product 3", Price = 30 });
            _products2.Add(new Product { IdProduct = 5, Description = "Series 2 - Product 5", Price = 230 });
            _products2.Add(new Product { IdProduct = 7, Description = "Series 2 - Product 7", Price = 240 });
            _products2.Add(new Product { IdProduct = 9, Description = "Series 2 - Product 9", Price = 250 });
            _products2.Add(new Product { IdProduct = 11, Description = "Series 2 - Product 11", Price = 260 });
        }

        private void btnSelect_Click(object sender, EventArgs e)
        {
            //this is a valid LINQ query but it doesn't show anything in the grid if your object does not have proper setter!!!
            //var result = from c in _customers
            //             select c;
            //lblResults.Text = "Rows in results = " + result.Count().ToString();
            //grdResults.DataSource = result.ToList();

            var result = from c in _customers
                         select new { CustomerName = c.Name, c.City, c.Country };
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnWhere_Click(object sender, EventArgs e)
        {
            //using real LINQ expression
            var result = from c in _customers
                         where c.City == "Montréal" || c.Name.ToUpper().Contains("JOE")
                         select new { CustomerName = c.Name, c.City, c.Country };
            ////using lambda expressions
            //var result = _customers
            //             .Where((c) => c.City == "Montréal" || c.Name.ToUpper().Contains("JOE"))
            //             .Select((c) => new { CustomerName = c.Name, c.City, c.Country });

            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnOrder_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         where c.Country == Countries.Canada
                         orderby c.Country, c.City, c.Name  
                         select new { c.Country, c.City, c.Name };
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnImplicitJoin_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         from o in c.Orders
                         select new { c.Country, c.City, c.Name, o.IdProduct, o.Quantity };
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnExplicitJoin_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         from o in c.Orders
                         join p in _products on o.IdProduct equals p.IdProduct
                         select new { c.Country, c.City, c.Name, o.IdProduct, o.Quantity, 
                                      p.Description, p.Price, Total = o.Quantity * p.Price };
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnLeftJoin_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         from o in c.Orders
                         join p in _products on o.IdProduct equals p.IdProduct into ProductsByOrder
                         from po in ProductsByOrder.DefaultIfEmpty(new Product
                                                                       {
                                                                           IdProduct = 0, 
                                                                           Description = "Unknown Product", 
                                                                           Price = 0
                                                                       })
                         select new { c.Country, c.City, c.Name, o.IdProduct, o.Quantity, 
                                      po.Description, po.Price, Total = o.Quantity * po.Price };
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnGroup_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         group c by c.Country into GroupedCustomers
                         select new { Country = GroupedCustomers.Key, 
                                      Count = GroupedCustomers.Count(), 
                                      GroupedCustomers };
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();

            foreach (var group in result)
            {
                Console.WriteLine(group.Country + " - " + group.Count);
                foreach (var detail in group.GroupedCustomers)
                {
                    Console.WriteLine("   " + detail.City + " - " + detail.Name);
                }
            }
        }

        private void btnDistinct_Click(object sender, EventArgs e)
        {
            var result = (from c in _customers
                          select new { c.Country }
                         ).Distinct();
            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnSetUnion_Click(object sender, EventArgs e)
        {
            var result = _products.Union(_products2);

            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnSetIntersect_Click(object sender, EventArgs e)
        {
            var result = _products.Intersect(_products2);

            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnSetExcept_Click(object sender, EventArgs e)
        {
            var result = _products.Except(_products2);
            //to have the full list of Except
            //var result = (_products.Except(_products2))
            //                .Union(_products2.Except(_products));

            lblResults.Text = "Rows in results = " + result.Count().ToString();
            grdResults.DataSource = result.ToList();
        }

        private void btnAggregate_Click(object sender, EventArgs e)
        {
            var agg = new List<string>
                {
                    "Count = " + _products.Count(),
                    "LongCount = " + _products.LongCount(),
                    "Sum Price = " + _products.Sum(p => p.Price),
                    "Average Price = " + _products.Average(p => p.Price),
                    "Min Price = " + _products.Min(p => p.Price),
                    "Max Price = " + _products.Max(p => p.Price)
                };

            var result = from a in agg
                         select new { Aggregate = a };

            lblResults.Text = "Rows in results = " + result.Count();
            grdResults.DataSource = result.ToList();
        }

        private void btnAggregate2_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         from o in c.Orders
                         join p in _products
                                on o.IdProduct equals p.IdProduct
                         select new { c.Name, o.IdOrder, TotalItem = o.Quantity * p.Price };

            var result2 = from c in _customers
                          join o in result
                              on c.Name equals o.Name
                          into CustomersOrders
                          select new { c.Name, 
                              NbOrders = CustomersOrders.Count(),
                              TotalOrders = CustomersOrders.Sum(o => o.TotalItem) };

            lblResults.Text = "Rows in results = " + result2.Count();
            grdResults.DataSource = result2.ToList();
        }

        private void btnAggregate3_Click(object sender, EventArgs e)
        {
            var result = from c in _customers
                         from o in c.Orders
                         join p in _products
                                on o.IdProduct equals p.IdProduct
                         select new { c.Name, o.IdOrder, TotalItem = o.Quantity * p.Price };

            var result2 = from c in _customers
                          join o in result
                              on c.Name equals o.Name
                          into CustomersOrders
                          select new
                          {
                              c.Name,
                              MaxOrders = CustomersOrders.Aggregate(0m, (currentValue, currentItem) =>
                                                          currentValue > currentItem.TotalItem ?
                                                          currentValue :
                                                          currentItem.TotalItem)
                          };

            //this can be used to find the Min value (notice the use of decimal.MaxValue)
             //var result2 = from c in _customers
             //             join o in result
             //                 on c.Name equals o.Name
             //             into CustomersOrders
             //             select new
             //             {
             //                 c.Name,
             //                 MinOrders = CustomersOrders.Aggregate(decimal.MaxValue, (accumulator, currentItem) =>
             //                                             accumulator < currentItem.TotalItem ?
             //                                             accumulator :
             //                                             currentItem.TotalItem)
             //             };
            lblResults.Text = "Rows in results = " + result2.Count();
            grdResults.DataSource = result2.ToList();
        }

        private void btnAggregate4_Click(object sender, EventArgs e)
        {
            var strFiles = System.IO.Directory.GetFiles(Application.StartupPath)
                .Aggregate((current, strFile) => current + (strFile + ";"));
            MessageBox.Show(strFiles);
        }

        private void btnRange_Click(object sender, EventArgs e)
        {
            var result = Enumerable.Range(1000, 100)
                .Select(x => (new Product
                {
                    IdProduct = x,
                    Description = "Product " + x,
                    Price = x * 2
                }));
            //join the result of Range and Product
            //var result = Enumerable.Range(1, 100)
            //.SelectMany(x => (from p in _products2 
            //                  where p.IdProduct == x 
            //                  select p));

            lblResults.Text = "Rows in results = " + result.Count();
            grdResults.DataSource = result.ToList();
        }

        private void btnRepeat_Click(object sender, EventArgs e)
        {
            var result = Enumerable.Repeat((from p in _products 
                                            where p.IdProduct == 1 
                                            select p), 7)
                                           .SelectMany(x => x);

            lblResults.Text = "Rows in results = " + result.Count();
            grdResults.DataSource = result.ToList();
        }

        private void btnAny_Click(object sender, EventArgs e)
        {
            bool result = (from c in _customers select c)
                .Any(c => c.City == "Montréal");
            if (result)
                MessageBox.Show("There is at least one customer in Montréal");
            else
                MessageBox.Show("There is no customer from Montréal");
        }

        private void btnAll_Click(object sender, EventArgs e)
        {
            bool result = (from c in _customers select c)
                .All(c => c.City == "Montréal");
            if (result)
                MessageBox.Show("All customers are from Montréal");
            else
                MessageBox.Show("NOT all customers are from Montréal");
        }

        private void btnContains_Click(object sender, EventArgs e)
        {
            var productToTest = new Product
            {
                IdProduct = 1,
                Description = "Product 1",
                Price = 10
            };
            bool result = _products.Contains(productToTest);
            if (result)
                MessageBox.Show("The list of products contains the tested element");
            else
                MessageBox.Show("The list of products DOES NOT contain the tested element");
        }

    }

    #region Types

    public enum Countries
    {
        Unknown,
        Canada,
        USA,
    }

    public class Customer
    {
        public string Name; // { get; set; }
        public string City; // { get; set; }
        public Countries Country; // { get; set; }
        public Order[] Orders; // { get; set; }

        //public override string ToString()
        //{
        //    return String.Format("Name: {0} - City: {1} - Country: {2}", Name, City, Country);
        //}

        //private static volatile Customer empty;
        //private static Object emptySyncLock = new Object();

        //public static Customer Empty
        //{
        //    get
        //    {
        //        // Multithreaded singleton pattern
        //        if (empty == null)
        //        {
        //            lock (emptySyncLock)
        //            {
        //                if (empty == null)
        //                {
        //                    empty = new Customer();
        //                    empty.Name = String.Empty;
        //                    empty.Country = Countries.Unknown;
        //                    empty.City = String.Empty;
        //                    empty.Orders = (new List<Order>(Enumerable.Empty<Order>())).ToArray();
        //                }
        //            }
        //        }
        //        return (empty);
        //    }
        //}
    }

    public class Order
    {
        public int IdOrder;
        public int Quantity;
        public bool Shipped;
        public string Month;
        public int IdProduct;

        //public override string ToString()
        //{
        //    return String.Format("IdOrder: {0} - IdProduct: {1} - Quantity: {2} - Shipped: {3} - Month: {4}",
        //        IdOrder, IdProduct, Quantity, Shipped, Month);
        //}

        //public override bool Equals(object obj)
        //{
        //    if (!(obj is Order))
        //        return false;
        //    else
        //    {
        //        Order o = (Order)obj;
        //        return (o.IdOrder == this.IdOrder && o.IdProduct == this.IdProduct && o.Month == this.Month && o.Quantity == this.Quantity && o.Shipped == this.Shipped);
        //    }
        //}

        //public override int GetHashCode()
        //{
        //    return String.Format("{0}|{1}|{2}|{3}|{4}", this.IdOrder, this.IdProduct, this.Month, this.Quantity, this.Shipped).GetHashCode();
        //}
    }

    public class Product
    {
        public int IdProduct { get; set; }
        public decimal Price { get; set; }
        public string Description { get; set; }

        //public override string ToString()
        //{
        //    return String.Format("IdProduct: {0} - Description {1} - Price: {2}", IdProduct, Description, Price);
        //}

        public override bool Equals(object obj)
        {
            if (!(obj is Product))
                return false;
            else
            {
                Product p = (Product)obj;
                return (p.IdProduct == IdProduct && p.Description == Description && p.Price == Price);
            }
        }

        public override int GetHashCode()
        {
            return String.Format("{0}|{1}|{2}", IdProduct, Description, Price).GetHashCode();
        }
    }

    #endregion
}
