﻿Option Strict On

Imports System.Windows.Forms

Public Class Form1

    Private _customers As New List(Of Customer)
    Private _products As New List(Of Product)
    Private _products2 As New List(Of Product)

    Private Sub Form1_Load(sender As Object, e As System.EventArgs) Handles Me.Load
        LoadData()
    End Sub

    Private Sub LoadData()
        _customers.Add(New Customer With {
                .Name = "Dalton, Joe",
                .City = "Toronto",
                .Country = Countries.Canada,
                .Orders = New Order() {
                    New Order With {.IdOrder = 1, .Quantity = 3, .IdProduct = 1, .Shipped = False, .Month = "January"},
                    New Order With {.IdOrder = 2, .Quantity = 5, .IdProduct = 2, .Shipped = True, .Month = "May"}}})
        _customers.Add(New Customer With {
                    .Name = "Dalton, Jack",
                    .City = "Montréal",
                    .Country = Countries.Canada,
                    .Orders = New Order() {
                        New Order With {.IdOrder = 3, .Quantity = 10, .IdProduct = 1, .Shipped = False, .Month = "July"},
                        New Order With {.IdOrder = 4, .Quantity = 20, .IdProduct = 3, .Shipped = True, .Month = "December"}}})
        _customers.Add(New Customer With {
                .Name = "Dalton, William",
                .City = "Winnipeg",
                .Country = Countries.Canada,
                .Orders = New Order() {
                    New Order With {.IdOrder = 5, .Quantity = 20, .IdProduct = 3, .Shipped = True, .Month = "December"}}})
        _customers.Add(New Customer With {
                .Name = "Dalton, Averell",
                .City = "Vancouver",
                .Country = Countries.Canada,
                .Orders = New Order() {
                    New Order With {.IdOrder = 6, .Quantity = 20, .IdProduct = 5, .Shipped = False, .Month = "July"}}})
        _customers.Add(New Customer With {
            .Name = "Moreau, Éric",
            .City = "Montréal",
            .Country = Countries.Canada,
            .Orders = New Order() {
                    New Order With {.IdOrder = 7, .Quantity = 33, .IdProduct = 1, .Shipped = True, .Month = "Juyly"},
                    New Order With {.IdOrder = 8, .Quantity = 55, .IdProduct = 2, .Shipped = False, .Month = "August"}}})
        _customers.Add(New Customer With {
            .Name = "US Customer 1",
            .City = "New York",
            .Country = Countries.USA,
            .Orders = New Order() {
                    New Order With {.IdOrder = 9, .Quantity = 7, .IdProduct = 5, .Shipped = True, .Month = "September"},
                    New Order With {.IdOrder = 10, .Quantity = 4, .IdProduct = 44, .Shipped = False, .Month = "October"}}
        })

        _products.Add(New Product With {.IdProduct = 1, .Description = "Product 1", .Price = 10})
        _products.Add(New Product With {.IdProduct = 2, .Description = "Product 2", .Price = 20})
        _products.Add(New Product With {.IdProduct = 3, .Description = "Product 3", .Price = 30})
        _products.Add(New Product With {.IdProduct = 4, .Description = "Product 4", .Price = 40})
        _products.Add(New Product With {.IdProduct = 5, .Description = "Product 5", .Price = 50})
        _products.Add(New Product With {.IdProduct = 6, .Description = "Product 6", .Price = 60})

        _products2.Add(New Product With {.IdProduct = 1, .Description = "Product 1", .Price = 10})
        _products2.Add(New Product With {.IdProduct = 3, .Description = "Product 3", .Price = 30})
        _products2.Add(New Product With {.IdProduct = 5, .Description = "Series 2 - Product 5", .Price = 230})
        _products2.Add(New Product With {.IdProduct = 7, .Description = "Series 2 - Product 7", .Price = 240})
        _products2.Add(New Product With {.IdProduct = 9, .Description = "Series 2 - Product 9", .Price = 250})
        _products2.Add(New Product With {.IdProduct = 11, .Description = "Series 2 - Product 11", .Price = 260})
    End Sub

    Private Sub btnSelect_Click(sender As System.Object, e As System.EventArgs) Handles btnSelect.Click
        'this is a valid LINQ query but it doesn't show anything in the grid if your object does not have proper setter!!!
        'Dim result = From c In _customers
        '             Select c 

        Dim result = From c In _customers
                     Select CustomerName = c.Name, c.City, c.Country
        'This Select would also be valid
        'Select New With {.CustomerName = c.Name, c.City, c.Country}
        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnWhere_Click(sender As System.Object, e As System.EventArgs) Handles btnWhere.Click
        'using real LINQ expression
        Dim result = From c In _customers
                     Where c.City = "Montréal" OrElse c.Name.ToUpper.Contains("JOE")
                     Select CustomerName = c.Name, c.City, c.Country
        'using lambda expressions
        'Dim result = _customers.
        '        Where(Function(c) c.City = "Montréal" OrElse c.Name.ToUpper.Contains("JOE")).
        '        Select(Function(c) New With {.CustomerName = c.Name, c.City, c.Country})

        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnOrder_Click(sender As System.Object, e As System.EventArgs) Handles btnOrder.Click
        Dim result = From c In _customers
                     Where c.Country = Countries.Canada
                     Order By c.Country, c.City, c.Name
                     Select c.Country, c.City, c.Name
        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnImplicitJoin_Click(sender As System.Object, e As System.EventArgs) Handles btnImplicitJoin.Click
        Dim result = From c In _customers
                     From o In c.Orders
                     Select c.Country, c.City, c.Name, o.IdProduct, o.Quantity

        'Shorter VB syntax. Do we really need this?
        'Dim result = From c In _customers, o In c.Orders
        '             Select c.Country, c.City, c.Name, o.IdProduct, o.Quantity

        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnExplicitJoin_Click(sender As System.Object, e As System.EventArgs) Handles btnExplicitJoin.Click
        Dim result = From c In _customers
                     From o In c.Orders
                     Join p In _products On o.IdProduct Equals p.IdProduct
                     Select c.Country, c.City, c.Name, o.IdProduct, o.Quantity,
                            p.Description, p.Price, Total = o.Quantity * p.Price
        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnLeftJoin_Click(sender As System.Object, e As System.EventArgs) Handles btnLeftJoin.Click
        Dim result = From c In _customers
                     From o In c.Orders
                     Group Join p In _products
                        On o.IdProduct Equals p.IdProduct
                        Into ProductsByOrder = Group
                     From po In ProductsByOrder.DefaultIfEmpty(
                        New Product With {
                              .IdProduct = 0,
                              .Description = "Unknown Product",
                              .Price = 0
                             })
                     Select c.Country, c.City, c.Name, o.IdProduct, o.Quantity,
                            po.Description, po.Price, Total = o.Quantity * po.Price
        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnGroup_Click(sender As System.Object, e As System.EventArgs) Handles btnGroup.Click
        Dim result = From c In _customers
                     Group c By c.Country Into GroupedCustomers = Group
                     Select Country, Count = GroupedCustomers.Count, GroupedCustomers
        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList

        For Each group In result
            Console.WriteLine(group.Country.ToString() + " - " + group.Count.ToString())
            For Each detail In group.GroupedCustomers
                Console.WriteLine("   " + detail.City + " - " + detail.Name)
            Next
        Next
    End Sub

    Private Sub btnDistinct_Click(sender As System.Object, e As System.EventArgs) Handles btnDistinct.Click
        Dim result = (From c In _customers
                      Select New With {Key c.Country}
                     ).Distinct()

        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnSetUnion_Click(sender As System.Object, e As System.EventArgs) Handles btnSetUnion.Click
        Dim result = _products.Union(_products2)

        lblResults.Text = "Rows in results = " + result.Count().ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnSetIntersect_Click(sender As System.Object, e As System.EventArgs) Handles btnSetIntersect.Click
        Dim result = _products.Intersect(_products2)

        lblResults.Text = "Rows in results = " + result.Count().ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnSetExcept_Click(sender As System.Object, e As System.EventArgs) Handles btnSetExcept.Click
        Dim result = _products.Except(_products2)
        'to have the full list of Except
        'Dim result = (_products.Except(_products2)).
        '    Union(_products2.Except(_products))

        lblResults.Text = "Rows in results = " + result.Count().ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnAggregate_Click(sender As System.Object, e As System.EventArgs) Handles btnAggregate.Click
        Dim agg = New List(Of String)

        agg.Add("Count = " + _products.Count().ToString())
        agg.Add("LongCount = " + _products.LongCount().ToString)
        agg.Add("Sum Price = " + _products.Sum(Function(p) p.Price).ToString)
        agg.Add("Average Price = " + _products.Average(Function(p) p.Price).ToString)
        agg.Add("Min Price = " + _products.Min(Function(p) p.Price).ToString)
        agg.Add("Max Price = " + _products.Max(Function(p) p.Price).ToString)

        Dim result = From a In agg
                     Select New With {.Aggregate = a}

        lblResults.Text = "Rows in results = " + result.Count.ToString
        grdResults.DataSource = result.ToList
    End Sub

    Private Sub btnAggregate2_Click(sender As System.Object, e As EventArgs) Handles btnAggregate2.Click
        Dim result = From c In _customers
             From o In c.Orders
             Join p In _products
                    On o.IdProduct Equals p.IdProduct
             Select c.Name, o.IdOrder, TotalItem = o.Quantity * p.Price

        Dim result2 = From c In _customers
                      Group Join o In result
                          On c.Name Equals o.Name
                      Into CustomersOrders = Group
                      Select c.Name,
                          NbOrders = CustomersOrders.Count(),
                          TotalOrders = CustomersOrders.Sum(Function(o) o.TotalItem)

        lblResults.Text = "Rows in results = " + result2.Count.ToString
        grdResults.DataSource = result2.ToList()
    End Sub

    Private Sub btnAggregate3_Click(sender As System.Object, e As EventArgs) Handles btnAggregate3.Click
        Dim result = From c In _customers
             From o In c.Orders
             Join p In _products
                    On o.IdProduct Equals p.IdProduct
             Select c.Name, o.IdOrder, TotalItem = o.Quantity * p.Price

        Dim result2 = From c In _customers
                      Group Join o In result
                          On c.Name Equals o.Name
                      Into CustomersOrders = Group
                      Select c.Name,
                              MaxOrders = CustomersOrders.Aggregate(0D, 
                                        Function(currentValue, currentItem) _
                                            If(currentValue > currentItem.TotalItem, 
                                               currentValue, 
                                               currentItem.TotalItem))

        ''this can be used to find the Min value (notice the use of decimal.MaxValue)
        'Dim result2 = From c In _customers
        '              Group Join o In result
        '                  On c.Name Equals o.Name
        '              Into CustomersOrders = Group
        '              Select c.Name,
        '                      MaxOrders = CustomersOrders.Aggregate(Decimal.MaxValue, 
        '                                Function(currentValue, currentItem) _
        '                                    If(currentValue < currentItem.TotalItem, 
        '                                       currentValue, 
        '                                       currentItem.TotalItem))

        lblResults.Text = "Rows in results = " + result2.Count.ToString
        grdResults.DataSource = result2.ToList()
    End Sub

    Private Sub btnAggregate4_Click(sender As System.Object, e As System.EventArgs) Handles btnAggregate4.Click
        Dim strFiles As String = IO.Directory.GetFiles(Application.StartupPath).
            Aggregate(strFiles, Function(current, strFile) current + (strFile + ";"))
        MessageBox.Show(strFiles)
    End Sub

    Private Sub btnRange_Click(sender As System.Object, e As System.EventArgs) Handles btnRange.Click
        Dim result = Enumerable.Range(1000, 100).
            Select(Function(x) (New Product With {
                                .IdProduct = x,
                                .Description = "Product " + x.ToString(),
                                .Price = x * 2
                            }))
        ''join the result of Range and Product
        'Dim result = Enumerable.Range(1, 100).
        '    SelectMany(Function(x) (
        '                   From p In _products2
        '                   Where p.IdProduct = x
        '                   Select p))

        lblResults.Text = "Rows in results = " + result.Count().ToString
        grdResults.DataSource = result.ToList()
    End Sub

    Private Sub btnRepeat_Click(sender As System.Object, e As System.EventArgs) Handles btnRepeat.Click
        Dim result = Enumerable.Repeat((From p In _products
                                        Where p.IdProduct = 1
                                        Select p), 7).
                            SelectMany(Function(x) x)

        lblResults.Text = "Rows in results = " + result.Count().ToString()
        grdResults.DataSource = result.ToList()
    End Sub

    Private Sub btnAny_Click(sender As System.Object, e As System.EventArgs) Handles btnAny.Click
        Dim result As Boolean = (From c In _customers Select c).
                                Any(Function(c) c.City = "Montréal")
        If (result) Then
            MessageBox.Show("There is at least one customer in Montréal")
        Else
            MessageBox.Show("There is no customer from Montréal")
        End If
    End Sub

    Private Sub btnAll_Click(sender As System.Object, e As System.EventArgs) Handles btnAll.Click
        Dim result As Boolean = (From c In _customers Select c).
                                All(Function(c) c.City = "Montréal")
        If (result) Then
            MessageBox.Show("All customers are from Montréal")
        Else
            MessageBox.Show("NOT all customers are from Montréal")
        End If
    End Sub

    Private Sub btnContains_Click(sender As System.Object, e As System.EventArgs) Handles btnContains.Click
        Dim productToTest = New Product With
        {
            .IdProduct = 1,
            .Description = "Product 1",
            .Price = 10
        }
        Dim result As Boolean = _products.Contains(productToTest)
        If (result) Then
            MessageBox.Show("The list of products contains the tested element")
        Else
            MessageBox.Show("The list of products DOES NOT contain the tested element")
        End If

    End Sub

End Class

Public Enum Countries
    Unknown
    Canada
    USA
End Enum

Public Class Customer
    Public Property Name As String
    Public Property City As String
    Public Property Country As Countries
    Public Property Orders As Order()

    'public function CityToString As string
    '    Return City
    'End Function

    'public Overrides function GetHashCode As Integer
    '    return City.GetHashCode()
    'End Function
End Class

Public Class Order
    Public IdOrder As Integer
    Public Quantity As Integer
    Public Shipped As Boolean
    Public Month As String
    Public IdProduct As Integer
End Class

Public Class Product
    Public Property IdProduct As Integer
    Public Property Price As Decimal
    Public Property Description As String

    Public Overrides Function Equals(obj As Object) As Boolean
        If obj Is Nothing Then Return False
        If Not (TypeOf obj Is Product) Then Return False

        Dim p As Product = DirectCast(obj, Product)
        Return (p.IdProduct = IdProduct AndAlso 
                p.Description = Description AndAlso 
                p.Price = Price)
    End Function

    Public Overrides Function GetHashCode() As Integer
        Return String.Format("{0}|{1}|{2}", IdProduct, Description, Price).GetHashCode
    End Function

End Class